@@ -1,4 +1,4 @@
-/* $scrotwm: scrotwm.c,v 1.107 2009/01/28 04:45:45 marco Exp $ */
+/* $scrotwm: scrotwm.c,v 1.123 2009/02/01 04:03:40 marco Exp $ */
 /*
  * Copyright (c) 2009 Marco Peereboom <marco@peereboom.us>
  * Copyright (c) 2009 Ryan McBride <mcbride@countersiege.com>
@@ -48,7 +48,9 @@
  * DEALINGS IN THE SOFTWARE.
  */
 
-#define	SWM_VERSION	"0.6"
+static const char	*cvstag = "$scrotwm: scrotwm.c,v 1.123 2009/02/01 04:03:40 marco Exp $";
+
+#define	SWM_VERSION	"0.7"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -102,7 +104,7 @@
 #define	SWM_D_STACK		0x0020
 #define	SWM_D_MOUSE		0x0040
 #define	SWM_D_PROP		0x0080
-
+#define	SWM_D_CLASS		0x0100
 
 u_int32_t		swm_debug = 0
 			    | SWM_D_MISC
@@ -111,7 +113,9 @@ u_int32_t		swm_debug = 0
 			    | SWM_D_FOCUS
 			    | SWM_D_MOVE
 			    | SWM_D_STACK
+			    | SWM_D_MOUSE
 			    | SWM_D_PROP
+			    | SWM_D_CLASS
 			    ;
 #else
 #define DPRINTF(x...)
@@ -122,18 +126,17 @@ u_int32_t		swm_debug = 0
 #define MODKEY			Mod1Mask
 #define CLEANMASK(mask)		(mask & ~(numlockmask | LockMask))
 #define BUTTONMASK		(ButtonPressMask|ButtonReleaseMask)
-
-#define X(r)		(r)->g.x	
-#define Y(r)		(r)->g.y
-#define WIDTH(r)	(r)->g.w	
-#define HEIGHT(r)	(r)->g.h
+#define MOUSEMASK		(BUTTONMASK|PointerMotionMask)
+#define SWM_PROPLEN		(16)
+#define X(r)			(r)->g.x	
+#define Y(r)			(r)->g.y
+#define WIDTH(r)		(r)->g.w	
+#define HEIGHT(r)		(r)->g.h
 
 #ifndef SWM_LIB
-#define SWM_LIB		"/usr/X11R6/lib/swmhack.so"
+#define SWM_LIB			"/usr/X11R6/lib/swmhack.so"
 #endif
 
-#define SWM_PROPLEN	(16)
-
 char			**start_argv;
 Atom			astate;
 int			(*xerrorxlib)(Display *, XErrorEvent *);
@@ -155,6 +158,8 @@ double			dialog_ratio = .6;
 char			*bar_argv[] = { NULL, NULL };
 int			bar_pipe[2];
 char			bar_ext[SWM_BAR_MAX];
+char			bar_vertext[SWM_BAR_MAX];
+int			bar_version = 0;
 sig_atomic_t		bar_alarm = 0;
 int			bar_delay = 30;
 int			bar_enabled = 1;
@@ -176,8 +181,8 @@ char			*bar_fonts[] = {
 /* terminal + args */
 char			*spawn_term[] = { "xterm", NULL };
 char			*spawn_screenshot[] = { "screenshot.sh", NULL, NULL };
-char			*spawn_menu[] = { "dmenu_run", "-fn", NULL,
-			    "-nb", NULL, "-nf", NULL, "-sb", NULL, "-sf", NULL, NULL };
+char			*spawn_menu[] = { "dmenu_run", "-fn", NULL, "-nb", NULL,
+			    "-nf", NULL, "-sb", NULL, "-sf", NULL, NULL };
 
 #define SWM_MENU_FN	(2)
 #define SWM_MENU_NB	(4)
@@ -213,10 +218,12 @@ struct ws_win {
 	int			got_focus;
 	int			floating;
 	int			transient;
+	int			manual;
 	struct workspace	*ws;	/* always valid */
 	struct swm_screen	*s;	/* always valid, never changes */
 	XWindowAttributes	wa;
 	XSizeHints		sh;
+	XClassHint		ch;
 };
 TAILQ_HEAD(ws_win_list, ws_win);
 
@@ -306,9 +313,54 @@ union arg {
 #define SWM_ARG_ID_CYCLESC_DOWN	(15)
 #define SWM_ARG_ID_SS_ALL	(0)
 #define SWM_ARG_ID_SS_WINDOW	(1)
+#define SWM_ARG_ID_DONTCENTER	(0)
+#define SWM_ARG_ID_CENTER	(1)
 	char			**argv;
 };
 
+/* quirks */
+struct quirk {
+	char			*class;
+	char			*name;
+	unsigned long		quirk;
+#define SWM_Q_FLOAT		(1<<0)
+} quirks[] = {
+	{ "MPlayer",		"xv",		SWM_Q_FLOAT },
+	{ "OpenOffice.org 2.4",	"VCLSalFrame",	SWM_Q_FLOAT },
+	{ NULL,		NULL,		0},
+};
+
+/* events */
+void			expose(XEvent *);
+void			keypress(XEvent *);
+void			buttonpress(XEvent *);
+void			configurerequest(XEvent *);
+void			configurenotify(XEvent *);
+void			destroynotify(XEvent *);
+void			enternotify(XEvent *);
+void			focusin(XEvent *);
+void			mappingnotify(XEvent *);
+void			maprequest(XEvent *);
+void			propertynotify(XEvent *);
+void			unmapnotify(XEvent *);
+void			visibilitynotify(XEvent *);
+
+void			(*handler[LASTEvent])(XEvent *) = {
+				[Expose] = expose,
+				[KeyPress] = keypress,
+				[ButtonPress] = buttonpress,
+				[ConfigureRequest] = configurerequest,
+				[ConfigureNotify] = configurenotify,
+				[DestroyNotify] = destroynotify,
+				[EnterNotify] = enternotify,
+				[FocusIn] = focusin,
+				[MappingNotify] = mappingnotify,
+				[MapRequest] = maprequest,
+				[PropertyNotify] = propertynotify,
+				[UnmapNotify] = unmapnotify,
+				[VisibilityNotify] = visibilitynotify,
+};
+
 unsigned long
 name_to_color(char *colorname)
 {
@@ -552,8 +604,8 @@ bar_update(void)
 	for (i = 0; i < ScreenCount(display); i++) {
 		x = 1;
 		TAILQ_FOREACH(r, &screens[i].rl, entry) {
-			snprintf(loc, sizeof loc, "%s     %d:%d    %s",
-			    s, x++, r->ws->idx + 1, bar_ext);
+			snprintf(loc, sizeof loc, "%s     %d:%d    %s    %s",
+			    s, x++, r->ws->idx + 1, bar_ext, bar_vertext);
 			bar_print(r, loc);
 		}
 	}
@@ -678,6 +730,17 @@ bar_setup(struct swm_region *r)
 	bar_refresh();
 }
 
+void
+version(struct swm_region *r, union arg *args)
+{
+	bar_version = !bar_version;
+	if (bar_version)
+		strlcpy(bar_vertext, cvstag, sizeof bar_vertext);
+	else
+		strlcpy(bar_vertext, "", sizeof bar_vertext);
+	bar_update();
+}
+
 void
 config_win(struct ws_win *win)
 {
@@ -804,6 +867,7 @@ void
 spawn(struct swm_region *r, union arg *args)
 {
 	char			*ret;
+	int			si;
 
 	DNPRINTF(SWM_D_MISC, "spawn: %s\n", args->argv[0]);
 	/*
@@ -824,6 +888,12 @@ spawn(struct swm_region *r, union arg *args)
 				free(ret);
 			}
 			setsid();
+			/* kill stdin, mplayer, ssh-add etc. need that */
+			si = open("/dev/null", O_RDONLY, 0);
+			if (si == -1)
+				err(1, "open /dev/null");
+			if (dup2(si, 0) == -1)
+				err(1, "dup2 /dev/null");
 			execvp(args->argv[0], args->argv);
 			fprintf(stderr, "execvp failed\n");
 			perror(" failed");
@@ -1169,8 +1239,13 @@ stack_floater(struct ws_win *win, struct swm_region *r)
 	}
 	wc.width = win->g.w;
 	wc.height = win->g.h;
-	wc.x = (WIDTH(r) - win->g.w) / 2;
-	wc.y = (HEIGHT(r) - win->g.h) / 2;
+	if (win->manual) {
+		wc.x = win->g.x;
+		wc.y = win->g.y;
+	} else {
+		wc.x = (WIDTH(r) - win->g.w) / 2;
+		wc.y = (HEIGHT(r) - win->g.h) / 2;
+	}
 
 	DNPRINTF(SWM_D_STACK, "stack_floater: win %lu x %d y %d w %d h %d\n",
 	    win->id, wc.x, wc.y, wc.width, wc.height);
@@ -1253,8 +1328,12 @@ stack_master(struct workspace *ws, struct swm_geometry *g, int rot, int flip)
 	extra = r_g.h - (colno * hrh);
 	win_g.h = hrh - 2;
 
+	/*  stack all the tiled windows */
 	i = j = 0;
 	TAILQ_FOREACH(win, &ws->winlist, entry) {
+		if (win->transient != 0 || win->floating != 0)
+			continue;
+
 		if (split && i == split) {
 			colno = winno - split;
 			hrh = (r_g.h / colno);
@@ -1295,35 +1374,40 @@ stack_master(struct workspace *ws, struct swm_geometry *g, int rot, int flip)
 		else
 			win_g.y += last_h + 2;
 
-		if (win->transient != 0 || win->floating != 0)
-			stack_floater(win, ws->r);
-		else {
-			bzero(&wc, sizeof wc);
-			wc.border_width = 1;
-			if (rot) {
-				win->g.x = wc.x = win_g.y;
-				win->g.y = wc.y = win_g.x;
-				win->g.w = wc.width = win_g.h;
-				win->g.h = wc.height = win_g.w;
-			} else {
-				win->g.x = wc.x = win_g.x;
-				win->g.y = wc.y = win_g.y;
-				win->g.w = wc.width = win_g.w;
-				win->g.h = wc.height = win_g.h;
-			}
-			mask = CWX | CWY | CWWidth | CWHeight | CWBorderWidth;
-			XConfigureWindow(display, win->id, mask, &wc);
-			/*
-			fprintf(stderr, "vertical_stack: win %d x %d y %d w %d h %d bw %d\n", win->id, win->g.x, win->g.y, win->g.w , win->g.h, wc.border_width);
-			*/
+		bzero(&wc, sizeof wc);
+		wc.border_width = 1;
+		if (rot) {
+			win->g.x = wc.x = win_g.y;
+			win->g.y = wc.y = win_g.x;
+			win->g.w = wc.width = win_g.h;
+			win->g.h = wc.height = win_g.w;
+		} else {
+			win->g.x = wc.x = win_g.x;
+			win->g.y = wc.y = win_g.y;
+			win->g.w = wc.width = win_g.w;
+			win->g.h = wc.height = win_g.h;
 		}
-
+		mask = CWX | CWY | CWWidth | CWHeight | CWBorderWidth;
+		XConfigureWindow(display, win->id, mask, &wc);
 		XMapRaised(display, win->id);
+		/*
+		fprintf(stderr, "vertical_stack: win %d x %d y %d w %d h %d bw %d\n", win->id, win->g.x, win->g.y, win->g.w , win->g.h, wc.border_width);
+		*/
+
 		last_h = win_g.h;
 		i++;
 		j++;
 	}
 
+	/* now, stack all the floaters and transients */
+	TAILQ_FOREACH(win, &ws->winlist, entry) {
+		if (win->transient == 0 && win->floating == 0)
+			continue;
+
+		stack_floater(win, ws->r);
+		XMapRaised(display, win->id);
+	}
+
 	if (winfocus)
 		focus_win(winfocus); /* has to be done outside of the loop */
 }
@@ -1461,6 +1545,9 @@ send_to_ws(struct swm_region *r, union arg *args)
 	Atom			ws_idx_atom = 0;
 	unsigned char		ws_idx_str[SWM_PROPLEN];
 
+	if (win == NULL)
+		return;
+
 	DNPRINTF(SWM_D_MOVE, "send_to_ws: win: %lu\n", win->id);
 
 	ws = win->ws;
@@ -1530,6 +1617,20 @@ screenshot(struct swm_region *r, union arg *args)
 	spawn(r, &a);
 }
 
+void
+floating_toggle(struct swm_region *r, union arg *args)
+{
+	struct ws_win	*win = cur_focus;
+
+	if (win == NULL)
+		return;
+
+	win->floating = !win->floating;
+	win->manual = 0;
+	stack();
+	focus_win(win);
+}
+
 /* key definitions */
 struct key {
 	unsigned int		mod;
@@ -1584,23 +1685,163 @@ struct key {
 	{ MODKEY | ShiftMask,	XK_x,		wkill,		{0} },
 	{ MODKEY,		XK_s,		screenshot,	{.id = SWM_ARG_ID_SS_ALL} },
 	{ MODKEY | ShiftMask,	XK_s,		screenshot,	{.id = SWM_ARG_ID_SS_WINDOW} },
+	{ MODKEY,		XK_t,		floating_toggle,{0} },
+	{ MODKEY | ShiftMask,	XK_v,		version,	{0} },
 };
 
 void
-click(struct swm_region *r, union arg *args)
+resize_window(struct ws_win *win, int center)
 {
-	DNPRINTF(SWM_D_MOUSE, "click: button: %d\n", args->id);
+	unsigned int		mask;
+	XWindowChanges		wc;
+	struct swm_region	*r;
 
-	switch (args->id) {
-	case Button1:
-		break;
-	case Button2:
-		break;
-	case Button3:
-		break;
-	default:
-		return;
+	r = root_to_region(win->wa.root);
+	bzero(&wc, sizeof wc);
+	mask = CWBorderWidth | CWWidth | CWHeight;
+	wc.border_width = 1;
+	wc.width = win->g.w;
+	wc.height = win->g.h;
+	if (center == SWM_ARG_ID_CENTER) {
+		wc.x = (WIDTH(r) - win->g.w) / 2;
+		wc.y = (HEIGHT(r) - win->g.h) / 2;
+		mask |= CWX | CWY;
 	}
+
+	DNPRINTF(SWM_D_STACK, "resize_window: win %lu x %d y %d w %d h %d\n",
+	    win->id, wc.x, wc.y, wc.width, wc.height);
+
+	XConfigureWindow(display, win->id, mask, &wc);
+	config_win(win);
+}
+
+void
+resize(struct ws_win *win, union arg *args)
+{
+	XEvent			ev;
+	Time			time = 0;
+
+	DNPRINTF(SWM_D_MOUSE, "resize: win %d floating %d trans %d\n",
+	    win->id, win->floating, win->transient);
+
+	if (!(win->transient != 0 || win->floating != 0))
+		return;
+
+	if (XGrabPointer(display, win->id, False, MOUSEMASK, GrabModeAsync,
+	    GrabModeAsync, None, None /* cursor */, CurrentTime) != GrabSuccess)
+		return;
+	XWarpPointer(display, None, win->id, 0, 0, 0, 0, win->g.w, win->g.h);
+	do {
+		XMaskEvent(display, MOUSEMASK | ExposureMask |
+		    SubstructureRedirectMask, &ev);
+		switch(ev.type) {
+		case ConfigureRequest:
+		case Expose:
+		case MapRequest:
+			handler[ev.type](&ev);
+			break;
+		case MotionNotify:
+			if (ev.xmotion.x < 0)
+				ev.xmotion.x = 0;
+			if (ev.xmotion.y < 0)
+				ev.xmotion.y = 0;
+			win->g.w = ev.xmotion.x;
+			win->g.h = ev.xmotion.y;
+
+			/* not free, don't sync more than 60 times / second */
+			if ((ev.xmotion.time - time) > (1000 / 60) ) {
+				time = ev.xmotion.time;
+				XSync(display, False);
+				resize_window(win, args->id);
+			}
+			break;
+		}
+	} while (ev.type != ButtonRelease);
+	if (time) {
+		XSync(display, False);
+		resize_window(win, args->id);
+	}
+	XWarpPointer(display, None, win->id, 0, 0, 0, 0, win->g.w - 1,
+	    win->g.h - 1);
+	XUngrabPointer(display, CurrentTime);
+
+	/* drain events */
+	while (XCheckMaskEvent(display, EnterWindowMask, &ev));
+}
+
+void
+move_window(struct ws_win *win)
+{
+	unsigned int		mask;
+	XWindowChanges		wc;
+	struct swm_region	*r;
+
+	r = root_to_region(win->wa.root);
+	bzero(&wc, sizeof wc);
+	mask = CWX | CWY;
+	wc.x = win->g.x;
+	wc.y = win->g.y;
+
+	DNPRINTF(SWM_D_STACK, "move_window: win %lu x %d y %d w %d h %d\n",
+	    win->id, wc.x, wc.y, wc.width, wc.height);
+
+	XConfigureWindow(display, win->id, mask, &wc);
+	config_win(win);
+}
+
+void
+move(struct ws_win *win, union arg *args)
+{
+	XEvent			ev;
+	Time			time = 0;
+	int			restack = 0;
+
+	DNPRINTF(SWM_D_MOUSE, "move: win %d floating %d trans %d\n",
+	    win->id, win->floating, win->transient);
+
+	if (win->floating == 0) {
+		win->floating = 1;
+		win->manual = 1;
+		restack = 1;
+	}
+
+	if (XGrabPointer(display, win->id, False, MOUSEMASK, GrabModeAsync,
+	    GrabModeAsync, None, None /* cursor */, CurrentTime) != GrabSuccess)
+		return;
+	XWarpPointer(display, None, win->id, 0, 0, 0, 0, 0, 0);
+	do {
+		XMaskEvent(display, MOUSEMASK | ExposureMask |
+		    SubstructureRedirectMask, &ev);
+		switch(ev.type) {
+		case ConfigureRequest:
+		case Expose:
+		case MapRequest:
+			handler[ev.type](&ev);
+			break;
+		case MotionNotify:
+			win->g.x = ev.xmotion.x_root;
+			win->g.y = ev.xmotion.y_root;
+
+			/* not free, don't sync more than 60 times / second */
+			if ((ev.xmotion.time - time) > (1000 / 60) ) {
+				time = ev.xmotion.time;
+				XSync(display, False);
+				move_window(win);
+			}
+			break;
+		}
+	} while (ev.type != ButtonRelease);
+	if (time) {
+		XSync(display, False);
+		move_window(win);
+	}
+	XWarpPointer(display, None, win->id, 0, 0, 0, 0, 0, 0);
+	XUngrabPointer(display, CurrentTime);
+	if (restack)
+		stack();
+
+	/* drain events */
+	while (XCheckMaskEvent(display, EnterWindowMask, &ev));
 }
 
 /* mouse */
@@ -1609,11 +1850,13 @@ struct button {
 	unsigned int		action;
 	unsigned int		mask;
 	unsigned int		button;
-	void			(*func)(struct swm_region *r, union arg *);
+	void			(*func)(struct ws_win *, union arg *);
 	union arg		args;
 } buttons[] = {
 	  /* action		key		mouse button	func		args */
-	{ client_click,		MODKEY,		Button1,	click, {.id=Button1} },
+	{ client_click,		MODKEY,		Button3,	resize, 	{.id = SWM_ARG_ID_DONTCENTER} },
+	{ client_click,		MODKEY | ShiftMask, Button3,	resize, 	{.id = SWM_ARG_ID_CENTER} },
+	{ client_click,		MODKEY,		Button1,	move, 		{0} },
 };
 
 void
@@ -1729,8 +1972,7 @@ buttonpress(XEvent *e)
 		if (action == buttons[i].action && buttons[i].func &&
 		    buttons[i].button == ev->button &&
 		    CLEANMASK(buttons[i].mask) == CLEANMASK(ev->state))
-			buttons[i].func(root_to_region(ev->root),
-			&buttons[i].args);
+			buttons[i].func(win, &buttons[i].args);
 }
 
 void
@@ -1750,14 +1992,12 @@ manage_window(Window id)
 	Window			trans;
 	struct workspace	*ws;
 	struct ws_win		*win;
-	XClassHint		ch;
-	int			format;
+	int			format, i, ws_idx;
 	unsigned long		nitems, bytes;
 	Atom			ws_idx_atom = 0, type;
 	unsigned char		ws_idx_str[SWM_PROPLEN], *prop = NULL;
 	struct swm_region	*r;
 	long			mask;
-	int			ws_idx;
 	const char		*errstr;
 
 	if ((win = find_window(id)) != NULL)
@@ -1766,45 +2006,49 @@ manage_window(Window id)
 	if ((win = calloc(1, sizeof(struct ws_win))) == NULL)
 		errx(1, "calloc: failed to allocate memory for new window");
 
+	/* Get all the window data in one shot */
 	ws_idx_atom = XInternAtom(display, "_SWM_WS", False);
 	if (ws_idx_atom)
 		XGetWindowProperty(display, id, ws_idx_atom, 0, SWM_PROPLEN,
 		    False, XA_STRING, &type, &format, &nitems, &bytes, &prop);
-
 	XGetWindowAttributes(display, id, &win->wa);
-	r = root_to_region(win->wa.root);
-	/* If the window was managed before, put it in the same workspace */
-	if (prop) {
-		DNPRINTF(SWM_D_PROP, "got property _SWM_WS=%s\n", prop);
-		ws_idx = strtonum(prop, 0, 9, &errstr);
-		if (errstr)
-			DNPRINTF(SWM_D_EVENT, "window idx is %s: %s",
-			    errstr, prop);
-		ws = &r->s->ws[ws_idx];
-	} else
-		ws = r->ws;
-
-	win->id = id;
-	win->ws = ws;
-	win->s = r->s;	/* this never changes */
-	TAILQ_INSERT_TAIL(&ws->winlist, win, entry);
-
-	/* make new win focused */
-	focus_win(win);
-
-	XGetTransientForHint(display, win->id, &trans);
+	XGetTransientForHint(display, id, &trans);
+	XGetWMNormalHints(display, id, &win->sh, &mask); /* XXX function? */
 	if (trans) {
 		win->transient = trans;
 		DNPRINTF(SWM_D_MISC, "manage_window: win %u transient %u\n",
 		    (unsigned)win->id, win->transient);
 	}
+	
+	/*
+	 * Figure out where to put the window. If it was previously assigned to
+	 * a workspace (either by spawn() or manually moving), and isn't
+	 * transient, * put it in the same workspace
+	 */
+	r = root_to_region(win->wa.root);
+	if (prop && win->transient == 0) {
+		DNPRINTF(SWM_D_PROP, "got property _SWM_WS=%s\n", prop);
+		ws_idx = strtonum(prop, 0, 9, &errstr);
+		if (errstr) {
+			DNPRINTF(SWM_D_EVENT, "window idx is %s: %s",
+			    errstr, prop);
+		}
+		ws = &r->s->ws[ws_idx];
+	} else
+		ws = r->ws;
+
+	/* set up the window layout */
+	win->id = id;
+	win->ws = ws;
+	win->s = r->s;	/* this never changes */
+	TAILQ_INSERT_TAIL(&ws->winlist, win, entry);
+
 	win->g.w = win->wa.width;
 	win->g.h = win->wa.height;
 	win->g.x = win->wa.x;
 	win->g.y = win->wa.y;
 
-	XGetWMNormalHints(display, win->id, &win->sh, &mask);
-
+	/* Set window properties so we can remember this after reincarnation */
 	if (ws_idx_atom && prop == NULL &&
 	    snprintf(ws_idx_str, SWM_PROPLEN, "%d", ws->idx) < SWM_PROPLEN) {
 		DNPRINTF(SWM_D_PROP, "setting property _SWM_WS to %s\n",
@@ -1818,17 +2062,19 @@ manage_window(Window id)
 	fprintf(stderr, "manage window: %d x %d y %d w %d h %d\n", win->id, win->g.x, win->g.y, win->g.w, win->g.h);
 	*/
 
-	/* XXX make this a table */
-	bzero(&ch, sizeof ch);
-	if (XGetClassHint(display, win->id, &ch)) {
-		/*fprintf(stderr, "class: %s name: %s\n", ch.res_class, ch.res_name); */
-		if (!strcmp(ch.res_class, "MPlayer") && !strcmp(ch.res_name, "xv")) {
-			win->floating = 1;
+	if (XGetClassHint(display, win->id, &win->ch)) {
+		DNPRINTF(SWM_D_CLASS, "class: %s name: %s\n",
+		    win->ch.res_class, win->ch.res_name);
+		for (i = 0; quirks[i].class != NULL && quirks[i].name != NULL &&
+		    quirks[i].quirk != 0; i++){
+			if (!strcmp(win->ch.res_class, quirks[i].class) &&
+			    !strcmp(win->ch.res_name, quirks[i].name)) {
+				DNPRINTF(SWM_D_CLASS, "found: %s name: %s\n",
+				    win->ch.res_class, win->ch.res_name);
+				if (quirks[i].quirk & SWM_Q_FLOAT)
+					win->floating = 1;
+			}
 		}
-		if (ch.res_class)
-			XFree(ch.res_class);
-		if (ch.res_name)
-			XFree(ch.res_name);
 	}
 
 	XSelectInput(display, id, EnterWindowMask | FocusChangeMask |
@@ -1836,6 +2082,9 @@ manage_window(Window id)
 
 	set_win_state(win, NormalState);
 
+	/* make new win focused */
+	focus_win(win);
+
 	return (win);
 }
 
@@ -1884,15 +2133,16 @@ configurerequest(XEvent *e)
 				win->g.h = ev->height;
 			if (win->ws->r != NULL) {
 				/* this seems to be full screen */
-				if (win->g.w > WIDTH(win->ws->r)) {
-					/* kill border */
-					win->g.x -= 1;
-					win->g.w += 1;
+				if (win->g.w >= WIDTH(win->ws->r)) {
+					win->g.x = -1;
+					win->g.w = WIDTH(win->ws->r);
+					ev->value_mask |= CWX | CWWidth;
 				}
-				if (win->g.h > HEIGHT(win->ws->r)) {
+				if (win->g.h >= HEIGHT(win->ws->r)) {
 					/* kill border */
-					win->g.y -= 1;
-					win->g.h += 1;
+					win->g.y = -1;
+					win->g.h = HEIGHT(win->ws->r);
+					ev->value_mask |= CWY | CWHeight;
 				}
 			}
 			if ((ev->value_mask & (CWX|CWY)) &&
@@ -1935,6 +2185,10 @@ destroynotify(XEvent *e)
 			focus_win(ws->focus);
 		TAILQ_REMOVE(&ws->winlist, win, entry);
 		set_win_state(win, WithdrawnState);
+		if (win->ch.res_class)
+			XFree(win->ch.res_class);
+		if (win->ch.res_name)
+			XFree(win->ch.res_name);
 		free(win);
 		stack();
 	}
@@ -1996,8 +2250,36 @@ maprequest(XEvent *e)
 void
 propertynotify(XEvent *e)
 {
+	struct ws_win		*win;
+	XPropertyEvent		*ev = &e->xproperty;
+
 	DNPRINTF(SWM_D_EVENT, "propertynotify: window: %lu\n",
-	    e->xproperty.window);
+	    ev->window);
+
+	if (ev->state == PropertyDelete)
+		return; /* ignore */
+	win = find_window(ev->window);
+	if (win == NULL)
+		return;
+
+	switch (ev->atom) {
+	case XA_WM_NORMAL_HINTS:
+#if 0
+		long		mask;
+		XGetWMNormalHints(display, win->id, &win->sh, &mask);
+		fprintf(stderr, "normal hints: flag 0x%x\n", win->sh.flags);
+		if (win->sh.flags & PMinSize) {
+			win->g.w = win->sh.min_width;
+			win->g.h = win->sh.min_height;
+			fprintf(stderr, "min %d %d\n", win->g.w, win->g.h);
+		}
+		XMoveResizeWindow(display, win->id,
+		    win->g.x, win->g.y, win->g.w, win->g.h);
+#endif
+		break;
+	default:
+		break;
+	}
 }
 
 void
@@ -2021,22 +2303,6 @@ visibilitynotify(XEvent *e)
 					bar_update();
 }
 
-void			(*handler[LASTEvent])(XEvent *) = {
-				[Expose] = expose,
-				[KeyPress] = keypress,
-				[ButtonPress] = buttonpress,
-				[ConfigureRequest] = configurerequest,
-				[ConfigureNotify] = configurenotify,
-				[DestroyNotify] = destroynotify,
-				[EnterNotify] = enternotify,
-				[FocusIn] = focusin,
-				[MappingNotify] = mappingnotify,
-				[MapRequest] = maprequest,
-				[PropertyNotify] = propertynotify,
-				[UnmapNotify] = unmapnotify,
-				[VisibilityNotify] = visibilitynotify,
-};
-
 int
 xerror_start(Display *d, XErrorEvent *ee)
 {
@@ -2263,7 +2529,8 @@ main(int argc, char *argv[])
 	fd_set			rd;
 
 	start_argv = argv;
-	fprintf(stderr, "Welcome to scrotwm V%s\n", SWM_VERSION);
+	fprintf(stderr, "Welcome to scrotwm V%s cvs tag: %s\n",
+	    SWM_VERSION, cvstag);
 	if (!setlocale(LC_CTYPE, "") || !XSupportsLocale())
 		warnx("no locale support");
 
